{-# LANGUAGE BangPatterns #-}
{-| Module      : Gargantext.Core.Viz.Graph.ProxemyOptim
Description : Proxemy
Copyright   : (c) CNRS, 2017-Present
License     : AGPL + CECILL v3
Maintainer  : team@gargantext.org
Stability   : experimental
Portability : POSIX

(/!\ To be published soon, confidential document for now. After
publication it will be integrated to the backend with usual license --
see above)

Article: Confluence for Graph Clustering, B. Gaume and A. Delanoë, Journal:TBA

Code written in Haskell by A. Delanoë from first Python Specifications by B.
Gaume.

-}

{-# LANGUAGE FlexibleContexts
           , DataKinds
           , GADTs
           , KindSignatures
           , ScopedTypeVariables
           , StandaloneDeriving
           , TypeOperators
           , TypeApplications
           , NoImplicitPrelude
           , RankNTypes
           , MultiParamTypeClasses
#-}

{-# LANGUAGE AllowAmbiguousTypes #-}
module Graph.BAC.ProxemyOptim
  where

import Data.IntMap (IntMap)
import Data.Maybe (isJust)
import Data.Proxy (Proxy(Proxy))
import Data.Reflection
import GHC.TypeLits (KnownNat, Nat, SomeNat(SomeNat), type(+), natVal, sameNat, someNatVal)
import Graph.FGL
import Graph.Types
import Prelude (String, readLn, error, id)
import Protolude hiding (traceShow, sum, natVal, trace)
import qualified Data.Graph.Inductive              as DGI
import qualified Data.Graph.Inductive.PatriciaTree as DGIP
import qualified Data.List                         as List
import qualified Data.Vector                       as V
import qualified Data.Vector.Storable              as VS
import qualified Data.Vector.Unboxed               as VU
import qualified Prelude                           as Prelude
import qualified Data.Set                          as Set

import qualified Data.Matrix.Sparse.Static         as SMatrix
import qualified Numeric.LinearAlgebra.Static      as DMatrix
import qualified Data.Vector.Unboxed as UV
import qualified Numeric.LinearAlgebra.Devel       as DMatrix
import qualified Data.Vector.Sparse.Static         as SVector
import qualified Data.IntMap.Strict as Dict
import qualified Data.Vector.Mutable as MV
import qualified Data.Vector.Unboxed.Mutable as MVU
import qualified Data.IntSet as IntSet
import Debug.Trace
import qualified Data.IntMap.Strict as IntMap

----------------------------------------------------------------

{-# INLINE clusteringOptim #-}
clusteringOptim :: forall n a b. KnownNat n
                => Length
                -> Similarity
                -> FiniteGraph n a b
                -> Double -- beta
                -> (Dict IntSet, Double)
clusteringOptim l s fg@(FiniteGraph g) beta = runClustering beta adj' prox sorted_edges
    where
      !adj' = symAdjacent fg True
      !tra' = symTransition adj'
      !prox = proxemie l tra'
      sorted_edges = sort_edges (natToInt @n) (edges_confluence l fg adj' tra')

      -- mc    = matconf False adj' prox
      -- md    = matmod fg

      -- !matq = case s of
      --   Conf -> SimConf mc
      --   Mod  -> SimMod  md

symAdjacent
  :: forall (n :: Nat) a b.
     KnownNat n
  => FiniteGraph n a b -> Bool -> SMatrix.Matrix n n Double
symAdjacent (FiniteGraph g) isReflexive = SMatrix.fromList (diag ++ triplets)
  where triplets = [ (i, j, 1.0) | i <- nodes g, j <- neighbors g i ]
        diag     = if isReflexive
                   then [ (i, i, 1.0) | i <- [0..(n-1)] ]
                   else []
        n        = fromIntegral $ natVal (Proxy :: Proxy n)

symTransition :: KnownNat n => SMatrix.Matrix n n Double -> SMatrix.Matrix n n Double
symTransition m = SMatrix.imap
  (\i j _ -> 1 / fromIntegral (SMatrix.nnzCol m j))
  m

-- | Where main Types are defined as
data Similarity = Conf | Mod
type Length      = Int

-- | A finite Graph is a Graph whose number of Nodes is a KnownNat
data FiniteGraph (n :: Nat) a b = FiniteGraph (Graph a b)

type IsReflexive = Bool
type NeighborsFilter a b = DGI.Gr a b -> Node -> [Node]
type RmEdge      = Bool
---------------------------------------------------------------
-- Data Structure
type VectorS          n = SVector.V n Double
type VectorD          n = DMatrix.L 1 n

type AdjacencyMatrix  n = MatrixS n
type TransitionMatrix n = MatrixS n
type ProxemyMatrix    n = MatrixS n

type ConfluenceMatrix n = MatrixD n
type ModularityMatrix n = MatrixD n

data SimilarityMatrix n = SimConf !(ConfluenceMatrix n)
                        | SimMod  !(ModularityMatrix n)

instance KnownNat n => Show (SimilarityMatrix n) where
  show (SimConf (DMatrix.L (DMatrix.Dim (DMatrix.Dim m)))) = show m
  show (SimMod m) = show m


-----
adjacent :: KnownNat n
         => FiniteGraph n a b
         -> IsReflexive
         -> AdjacencyMatrix n
adjacent (FiniteGraph g) isReflexive =
  SMatrix.fromList $ triplets <> diag
  where
    triplets = [ (i, j, 1.0)
               | i <- nodes g
               , j <- neighbors g i
               , i /= j
               ]

    diag = case isReflexive of
      True  -> [ (n, n, 1.0) | n <- nodes g ]
      False -> []


transition :: KnownNat n
           => AdjacencyMatrix  n
           -> TransitionMatrix n
transition m = SMatrix.imap
  (\i j _ -> 1 / fromIntegral (SMatrix.nnzCol m j))
  m

proxemie :: KnownNat n
         => Length
         -> SMatrix.Matrix n n Double
         -> ProxemyMatrix n
proxemie l !tm = trace "proxemie" $ case l <= 1 of
  True  -> tm
  False -> case iterate (SMatrix.mul tm) tm Prelude.!! (l-1) of
    tm' -> trace "proxemie OK" tm'

---------------------------------------------------------------
matconf :: forall n. KnownNat n
        => RmEdge
        -> SMatrix.Matrix n n Double
        -> ProxemyMatrix    n
        -> ConfluenceMatrix n
matconf False a p = seq a $ seq p $ trace "matconf" confmat
  where
    vcount  = natToInt @n
    sumdeg  = fromIntegral (SMatrix.nonZeros a)
    prox_nnz = 100 * fromIntegral (SMatrix.nonZeros p) / fromIntegral (vcount * vcount)
    p' = SMatrix.densify p
    confmat = DMatrix.L . DMatrix.Dim . DMatrix.Dim $ DMatrix.runSTMatrix $ do
      m <- DMatrix.newMatrix 0 vcount vcount
      forM_ [0..(vcount-1)] $ \x -> do
        let deg_x = fromIntegral (SMatrix.nnzCol a x)
        forM_ [(x+1)..(vcount-1)] $ \y -> do
            let prox_y_x_length = p' `DMatrix.at` (x,y)
                prox_y_x_infini = fromIntegral (SMatrix.nnzCol a x) / sumdeg
                conf = (prox_y_x_length - prox_y_x_infini) / (prox_y_x_length + prox_y_x_infini)
            DMatrix.unsafeWriteMatrix m x y conf
            DMatrix.unsafeWriteMatrix m y x conf
      trace "matconf OK" $ return m
matconf True _a _p = panic "MatConf True: TODO but not needed for now"

confAt
  :: forall (n :: Nat).
     KnownNat n
  => Double -- ^ beta
  -> SMatrix.Matrix n n Double -- ^ adjacency
  -> SMatrix.Matrix n n Double -- ^ proxemie
  -> Int -> Int -> Double
confAt beta adj prox x y = conf
  where deg_x = fromIntegral (SMatrix.nnzCol adj x)
        deg_y = fromIntegral (SMatrix.nnzCol adj y)
        sumdeg = fromIntegral (SMatrix.nonZeros adj)
        vcount = fromIntegral $ natVal (Proxy :: Proxy n)
        ecount = sumdeg - vcount
        -- ecount = 2 * number of edges, a bitof a misnomer
        -- indeed, sumdeg = number of non-zero entries in (reflexive) adj matrix,
        -- and '- vcount' removes the diagonal, giving 2 * |E|
        prox_y_x_length = prox `SMatrix.at` (x, y)
        prox_y_x_infini = deg_x / sumdeg
        conf = (prox_y_x_length - prox_y_x_infini) / (prox_y_x_length + prox_y_x_infini)
        xy | adj `SMatrix.at` (x, y) > 0 =
               conf + (1 - (deg_x-1)*(deg_y-1)/ecount)
           | otherwise =
               conf - (deg_x-1)*(deg_y-1)/ecount - beta*(1-conf)

matmod :: forall n a b. KnownNat n => FiniteGraph n a b -> ModularityMatrix n
matmod fg = DMatrix.L . DMatrix.Dim . DMatrix.Dim $ DMatrix.runSTMatrix go
  where
    !vcount = natToInt @n
    !a = adjacent fg False
    nnz = SMatrix.nonZeros a
    go :: forall s. ST s (DMatrix.STMatrix s Double)
    go = do
      m <- DMatrix.newMatrix 0 vcount vcount
      forM_ [0..(vcount-1)] $ \i ->
        forM_ [i..(vcount-1)] $ \j -> do
          let !v = if SMatrix.at a (i, j) > 0
                   then 1
                   else 0
              !v' =
                v - ( fromIntegral (SMatrix.nnzCol a i)
                    * fromIntegral (SMatrix.nnzCol a j)
                    / fromIntegral (2 * nnz)
                    )
          DMatrix.unsafeWriteMatrix m i j v'
          DMatrix.unsafeWriteMatrix m j i v'
      return m

---------------------------------------------------------------
type UnsortedEdges = [(Node, Node, Double)]
type SortedEdges   = [(Node, Node, Double)]

type X = Edge

edges_confluence :: forall n a b.
                    KnownNat n
                 => Length
                 -> FiniteGraph n a b
                 -> SMatrix.Matrix n n Double -- adjacency
                 -> SMatrix.Matrix n n Double -- transition
                 -> UnsortedEdges
edges_confluence l (FiniteGraph g) am tm = trace "edges_confluence" $
  map f (edges g)

  where
      vcount    = natToInt @n
      sumdeg_m2 = fromIntegral (SMatrix.nonZeros am - 2)

      f (x, y) = -- traceShow (x, y) $
          let !deg_x_m1 = fromIntegral (SMatrix.nnzCol am x - 1)
              !deg_y_m1 = fromIntegral (SMatrix.nnzCol am y - 1)
              v         = SMatrix.asColumn (SVector.singleton y 1)
              v'        =
                SMatrix.withColChangeExcept x (1/deg_x_m1) y tm $ \tm' ->
                  SMatrix.withColChangeExcept y (1/deg_y_m1) x tm' $ \tm'' ->
                    iterate (SMatrix.mul tm'') v Prelude.!! 3
              prox_y_x_length = SMatrix.extractCol v' 0 SVector.! x
              prox_y_x_infini = deg_x_m1 / sumdeg_m2
              conf  = (prox_y_x_length - prox_y_x_infini) / (prox_y_x_length + prox_y_x_infini)
          in seq conf (x, y, conf)

      pp_vec dv@(DMatrix.R (DMatrix.Dim v)) =
        let strs = map show (VS.toList v)
            maxl = maximum (map length strs)
            padr m s = s ++ replicate (m - length s) ' '
        in intercalate "\n" (map (padr maxl) strs)
      pp_mul m v =
        let mats = Prelude.lines $ SMatrix.pp m
            vecs = Prelude.lines $ pp_vec v
        in trace (("---\n" ++) . Prelude.unlines $ Prelude.zipWith (\a b -> a ++ " | " ++ b) mats vecs ++ ["---"]) (SMatrix.mulV m v)

sort_edges :: Int
           -> UnsortedEdges
           -> SortedEdges
sort_edges n = List.sortBy (\a b -> confCompare a b <> comparing xnpy a b)
             -- . List.filter (\(x,y,_) -> x < y)
  where
    third
      :: forall a b c
       . (a,b,c) -> c
    third (_,_,c) = c
    xnpy (x,y,_) = x*n+y
    confCompare a b =
      if abs (third b - third a) < 10**(-12)
        then EQ
        else comparing (Down . third) a b


---------------------------------------------------------------

data PartData = PartData
  { partElems :: !IntSet
  , partScore :: {-# UNPACK #-} !Double
  } deriving (Show, Eq)

scoreMerge :: (Node -> Node -> Double) -> Maybe PartData -> Maybe PartData -> Double
scoreMerge f Nothing  Nothing  = 0
scoreMerge f Nothing  (Just p) = partScore p
scoreMerge f (Just p) Nothing  = partScore p
scoreMerge f (Just p) (Just q) =
  IntSet.foldl'
    (\s i -> IntSet.foldl' (\s' j -> s' + f i j) s (partElems p))
    (partScore p + partScore q)
    (partElems q)

partsMerge :: Maybe PartData -> Maybe PartData -> Double -> Maybe PartData
partsMerge Nothing  Nothing  _ = Nothing
partsMerge Nothing  (Just p) _ = Just p
partsMerge (Just p) Nothing  _ = Just p
partsMerge (Just p) (Just q) s = Just $
  PartData (partElems p `IntSet.union` partElems q) s

data MClustering s =
  MClustering { mparts :: V.MVector s (Maybe PartData)
              , mindex :: VU.MVector s Int
              , mscore :: VU.MVector s Double -- 1-entry array, total score
              , mnumcl :: VU.MVector s Int    -- 1-entry array, #clusters
              -- TODO: mode?
              }

newMClustering :: Int -> ST s (MClustering s)
newMClustering n = do
  mps <- MV.unsafeNew n
  mis <- MVU.unsafeNew n
  msc <- MVU.unsafeNew 1
  mcl <- MVU.unsafeNew 1
  MVU.unsafeWrite msc 0 0.0
  MVU.unsafeWrite mcl 0 n
  forM_ [0..(n-1)] $ \i -> do
    MV.unsafeWrite mps i $ Just $ PartData (IntSet.singleton i) 0
    MVU.unsafeWrite mis i i
  return (MClustering mps mis msc mcl)

clusteringStep
  :: KnownNat n
  => Double -- ^ beta
  -> SMatrix.Matrix n n Double -- ^ adjacency
  -> SMatrix.Matrix n n Double -- ^ proxemie
  -> MClustering s -> (Node, Node) -> ST s ()
clusteringStep beta adj prox mclust (x, y) = do
  modX <- MVU.unsafeRead (mindex mclust) x
  modY <- MVU.unsafeRead (mindex mclust) y
  when (x /= y && modX /= modY) $ do
    partX <- MV.unsafeRead (mparts mclust) modX
    partY <- MV.unsafeRead (mparts mclust) modY
    score <- MVU.unsafeRead (mscore mclust) 0
    let scoreX  = maybe 0 partScore partX
        scoreY  = maybe 0 partScore partY
        scoreXY = scoreMerge f partX partY
        partXY  = partsMerge partX partY scoreXY
        score'  = score - scoreX - scoreY + scoreXY
    when (score' >= score) $ do
      MV.unsafeWrite (mparts mclust) modX partXY
      MV.unsafeWrite (mparts mclust) modY Nothing
      MVU.unsafeWrite (mscore mclust) 0 score'
      MVU.unsafeModify (mnumcl mclust) pred 0
      forM_ (partElems <$> partY) $ \ys ->
        IntSet.foldr (\i m -> MVU.unsafeWrite (mindex mclust) i modX >> m)
                     (return ())
                     ys

  where f x y = 2 * confAt beta adj prox x y

clusteringCollector
  :: forall s (n :: Nat).
     KnownNat n
  => Double -- ^ beta
  -> SMatrix.Matrix n n Double -- ^ adjacency
  -> SMatrix.Matrix n n Double -- ^ proxemie
  -> MClustering s
  -> ST s (Dict IntSet, Double)
clusteringCollector beta adj prox mclust = do
  nclust <- trace "counting non empty clusters" $ MV.foldl'
    (\n_acc mpart ->
       if isNothing mpart
       then n_acc
       else 1+n_acc
    )
    0
    (mparts mclust)
  newClusts <- trace ("creating new cluster vector of size " ++ show nclust) $ MV.unsafeNew nclust
  let go new_i _old_i Nothing  = return new_i
      go new_i old_i (Just p) = do
        MV.unsafeWrite newClusts new_i (Just p)
        return (new_i+1)
  MV.ifoldM' go 0 (mparts mclust)
  mat_delta <- trace "creating delta matrix" $ DMatrix.newMatrix (negate maxDouble) nclust nclust
  trace "filling delta matrix" $ forM_ [0..(nclust-1)] $ \i ->
    forM_ [(i+1)..(nclust-1)] $ \j -> do
      DMatrix.unsafeWriteMatrix mat_delta i j 0
      part_i <- MV.unsafeRead newClusts i
      part_j <- MV.unsafeRead newClusts j
      forPart part_i $ \x ->
        forPart part_j $ \y ->
          DMatrix.modifyMatrix mat_delta i j $ \a -> a + confAt beta adj prox x y
  delta0 <- trace "reading current score" $ MVU.unsafeRead (mscore mclust) 0
  let clusts = IntSet.fromList [0..(nclust-1)]
      argmaxRow i = do
        foldM (\acc@(max_j, !max_v) j -> do
                  v <- DMatrix.unsafeReadMatrix mat_delta i j
                  if v > max_v then return (j, v) else return acc
              )
              (-1, negate maxDouble)
              [(i+1)..(nclust-1)]
      bestPair cs = fmap (maximumBy (comparing (\(_, _, c) -> c))) $
        forM (IntSet.toAscList cs) $ \i ->
          (\(j, v) -> (i, j, v)) <$> argmaxRow i
      fusionRound !cs !merges !delta = do
        (good_i, good_j, delta') <- bestPair cs
        if delta' > 0 then do
          case IntSet.split good_i cs of
            (before_i, after_i) -> do
              forM_ (IntSet.toList before_i) $ \i -> do
                delta_ij <- DMatrix.unsafeReadMatrix mat_delta i good_j
                DMatrix.modifyMatrix mat_delta i good_i (+ delta_ij)
              forM_ (IntSet.toList after_i) $ \i -> do
                let (x, y) = if i < good_j then (i, good_j) else (good_j, i)
                when (x < y) $ do
                  delta_xy <- DMatrix.unsafeReadMatrix mat_delta x y
                  DMatrix.modifyMatrix mat_delta good_i i (+ delta_xy)
              case IntSet.split good_j cs of
                (before_j, _) -> forM_ (IntSet.toList before_j) $ \i ->
                  DMatrix.unsafeWriteMatrix mat_delta i good_j (negate maxDouble)
          let merges' = IntMap.insertWith IntSet.union good_i
                          (IntSet.insert good_j $ fromMaybe IntSet.empty $ IntMap.lookup good_j merges)
                          merges
          fusionRound (IntSet.delete good_j cs)
                      merges'
                      (delta + delta')
        else return (cs, merges, delta)
  (cs, merges, finalDelta) <- trace "cluster fusion" $ fusionRound clusts IntMap.empty delta0
  let groups = [ (i, maybe [] IntSet.toList (IntMap.lookup i merges) )
               | i <- IntSet.toList cs
               ]
  clustsDict <- trace "preparing collector results" . fmap IntMap.unions . forM (IntSet.toList cs) $ \i ->
    IntMap.singleton i . maybe IntSet.empty partElems <$> MV.unsafeRead newClusts i
  c <- foldGroups groups clustsDict $ \dict (i, js) -> do
    sets_i <- traverse (fmap (maybe IntSet.empty partElems) . MV.unsafeRead newClusts) js
    return $! IntMap.insertWith IntSet.union i (IntSet.unions sets_i) dict
  return (c, finalDelta)

  where maxDouble = encodeFloat m k :: Double
        b = floatRadix  (0 :: Double)
        e = floatDigits (0 :: Double)
        (_, e') = floatRange (0 :: Double)
        m = b^e - 1
        k = e' - e
        n = fromIntegral $ natVal (Proxy :: Proxy n)
        foldGroups gs cs f = foldM f cs gs
        forPart Nothing _  = return ()
        forPart (Just p) f =
          IntSet.foldr (\i acc -> f i >> acc) (return ()) (partElems p)

data Clust = Clust
  { cparts :: !(Dict IntSet)
  , cindex :: !(VU.Vector Int)
  , cscore :: {-# UNPACK #-} !Double
  } deriving (Show, Eq)

runClustering
  :: forall (n :: Nat). KnownNat n
  => Double -- ^ beta
  -> SMatrix.Matrix n n Double -- ^ adjacency
  -> SMatrix.Matrix n n Double -- ^ proxemie
  -> SortedEdges
  -> (Dict IntSet, Double)
runClustering beta adj prox se = trace "clustering starts" $ runST $ do
  mclust <- newMClustering n
  forM_ se $ \(x, y, _) -> clusteringStep beta adj prox mclust (x, y)
  trace "basic clusters done, running collector now" $ clusteringCollector beta adj prox mclust

  where n = fromIntegral $ natVal (Proxy :: Proxy n)
        sestr = intercalate "\n"
          [ show x ++ " " ++ show y ++ " " ++ show c
          | (x, y, c) <- se
          ]

-- the code below is unused for now

data Clustering a
  = ClusteringIs { parts :: !(Dict (Set a))
                 , index :: !(Dict Int)
                 , score :: !(Double)
                 , mode  :: !ClusteringMode
                 }
  | Clusterings { strict :: !(StrictClustering   a)
                , over   :: !(OverlapsClustering a)
                }
  deriving (Show, Eq)

data ClusteringMode = Part | Over | Both
  deriving (Show, Eq)

type StrictClustering   a = Clustering a
type OverlapsClustering a = Clustering a

make_clust_part :: forall n
                 . KnownNat n
                => SortedEdges
                -> SimilarityMatrix n
                -> Clustering Node
make_clust_part se sm =
  foldl' (\c (e1,e2,_) ->
            updateClustering c
              (\x y -> 2 * (sm' `DMatrix.at` (x, y)))
              e1 e2
         )
         (ClusteringIs parts idx 0 Part) se

  where
    ns    = [0 .. natToInt @n - 1 ]
    parts = Dict.fromAscList $ List.map (\n -> (n, Set.singleton n)) ns
    idx   = Dict.fromAscList (List.zip ns ns)
    sm'   = case sm of
      SimConf cm -> cm
      SimMod  mm -> mm

updateClustering :: Clustering Node
                 -> (Int -> Int -> Double)
                 -> Int -> Int -> Clustering Node
updateClustering c@(ClusteringIs parts idx currentScore _) f x y =
  let modX = fromMaybe (error "modX") $ Dict.lookup x idx
      modY = fromMaybe (error "modY") $ Dict.lookup y idx
   in case x == y || modX == modY of
    True  -> c -- do case x' or y' are Nothing
    False -> let c' = updateWith c f x modX y modY
              in case score c' >= currentScore of
                   True  -> c'
                   False -> c

updateWith :: Clustering Node
           -> (Int -> Int -> Double)
           -> Int -> Int
           -> Int -> Int
           -> Clustering Node
updateWith c@(ClusteringIs parts idx _ _) f x modX y modY =
  ClusteringIs parts' idx' score' Part
    where
      parts' = Dict.alter (add y) modX
             $ Dict.alter (del y) modY parts

      add y Nothing   = Just (Set.singleton y)
      add y (Just y') = Just (Set.insert y y')

      del y Nothing   = Nothing
      del y (Just y') = case Set.delete y y' of
        s | Set.null s -> Nothing
          | otherwise  -> Just s

      idx'         = Dict.alter (alter modX) y idx
      alter my _   = Just my

      score' = getSum $
        foldMap (\s -> fold [ Sum (f x'' y'')
                            | x'' <- Set.toList s
                            , y'' <- Set.toList s
                            , x'' < y''
                            ]
                ) parts'

---------------------------------------------------------------

------------------------------
make_clust_over :: KnownNat n
                => SimilarityMatrix n
                -> StrictClustering   a
                -> OverlapsClustering a
make_clust_over = Prelude.undefined

---------------------------------------------------------------
-- | Some Utils

-- | The Main Trick: Graph must be finite, then n can be known 1 time at least
--  The same n is needed the size of the Graph and Matrixes and Vectors
--  Proof by compilation of the compatibilities of the calculations
--  Then Proof by tests is less needed at this specific step of the flow.
buildFiniteGraph :: Proxy n -> Graph a b -> FiniteGraph n a b
buildFiniteGraph _ g = FiniteGraph g

-- | To execute a function on finite Graph
-- the KnownNat n is computed one time only
withG :: (Show a, Show b)
      => Graph a b
      -> (forall n. KnownNat n => FiniteGraph n a b -> r)
      -> r
withG g f = reifyNat (fromIntegral $ length $ nodes g)
          $ \n -> f (buildFiniteGraph n g)

natToInt :: forall (n :: Nat). KnownNat n => Int
natToInt = fromIntegral $ natVal (Proxy :: Proxy n)
